from __future__ import annotations

import json

from lazyagent.agent_observers import (
    AgentLifecycleEvent,
    AgentObserver,
    ClaudeHooksObserver,
    CodexAppServerObserver,
    CompositeObserver,
    GeminiTelemetryObserver,
    GeminiPromptObserver,
    LifecycleConfidence,
)
from lazyagent.models import AgentStatus


class TestCompositeObserver:
    def test_flattens_events_from_multiple_observers(self):
        class FakeObserver(AgentObserver):
            def __init__(self, event):
                self._event = event

            def on_process_started(self):
                return [self._event]

        evt_a = AgentLifecycleEvent(
            status=AgentStatus.RUNNING,
            confidence=LifecycleConfidence.HIGH,
            detail="a",
        )
        evt_b = AgentLifecycleEvent(
            status=AgentStatus.WAITING,
            confidence=LifecycleConfidence.MEDIUM,
            detail="b",
        )
        composite = CompositeObserver([FakeObserver(evt_a), FakeObserver(evt_b)])
        events = composite.on_process_started()
        assert len(events) == 2
        assert events[0] is evt_a
        assert events[1] is evt_b

    def test_cleanup_calls_all_observers(self):
        cleaned = []

        class TrackingObserver(AgentObserver):
            def __init__(self, name):
                self._name = name

            def cleanup(self):
                cleaned.append(self._name)

        composite = CompositeObserver([TrackingObserver("a"), TrackingObserver("b")])
        composite.cleanup()
        assert cleaned == ["a", "b"]


class TestGeminiPromptObserver:
    def test_prompt_visible_sets_waiting(self):
        observer = GeminiPromptObserver()
        events = observer.on_screen_update(
            "some output\n\ngemini > ",
            current_status=AgentStatus.RUNNING,
            current_detail="",
        )
        assert len(events) == 1
        assert events[0].status == AgentStatus.WAITING
        assert events[0].detail == "gemini prompt detected on screen"

    def test_prompt_missing_resumes_running(self):
        observer = GeminiPromptObserver()
        events = observer.on_screen_update(
            "some output\n\nrunning command...",
            current_status=AgentStatus.WAITING,
            current_detail="gemini prompt detected on screen",
        )
        assert len(events) == 1
        assert events[0].status == AgentStatus.RUNNING

    def test_no_transition_if_detail_mismatch(self):
        observer = GeminiPromptObserver()
        # If someone else set WAITING, we don't force it back to RUNNING
        events = observer.on_screen_update(
            "some output\n\nno prompt here",
            current_status=AgentStatus.WAITING,
            current_detail="sentinel visible on rendered screen",
        )
        assert events == []


class TestClaudeHooksObserver:
    def test_notification_event_sets_waiting_for_approval(self, tmp_path):
        log_path = tmp_path / "hooks.jsonl"
        log_path.write_text(
            json.dumps(
                {
                    "hook_event_name": "Notification",
                    "notification_type": "permission_prompt",
                }
            )
            + "\n",
            encoding="utf-8",
        )
        observer = ClaudeHooksObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.WAITING_FOR_APPROVAL
        assert events[0].confidence == LifecycleConfidence.HIGH

    def test_idle_prompt_sets_waiting_for_user(self, tmp_path):
        log_path = tmp_path / "hooks.jsonl"
        log_path.write_text(
            json.dumps(
                {
                    "hook_event_name": "Notification",
                    "notification_type": "idle_prompt",
                }
            )
            + "\n",
            encoding="utf-8",
        )
        observer = ClaudeHooksObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.WAITING_FOR_USER
        assert events[0].detail == "idle"

    def test_elicitation_sets_waiting_for_user(self, tmp_path):
        log_path = tmp_path / "hooks.jsonl"
        log_path.write_text(
            json.dumps(
                {
                    "hook_event_name": "Notification",
                    "notification_type": "elicitation_dialog",
                }
            )
            + "\n",
            encoding="utf-8",
        )
        observer = ClaudeHooksObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.WAITING_FOR_USER
        assert events[0].detail == "asking a question"

    def test_stop_event_sets_waiting(self, tmp_path):
        log_path = tmp_path / "hooks.jsonl"
        log_path.write_text(
            json.dumps({"hook_event_name": "Stop"}) + "\n",
            encoding="utf-8",
        )
        observer = ClaudeHooksObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.WAITING

    def test_session_end_sets_completed(self, tmp_path):
        log_path = tmp_path / "hooks.jsonl"
        log_path.write_text(
            json.dumps({"hook_event_name": "SessionEnd"}) + "\n",
            encoding="utf-8",
        )
        observer = ClaudeHooksObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.COMPLETED
        assert events[0].confidence == LifecycleConfidence.HIGH
        assert events[0].detail == "session ended"

    def test_pre_tool_use_sets_running(self, tmp_path):
        log_path = tmp_path / "hooks.jsonl"
        log_path.write_text(
            json.dumps(
                {"hook_event_name": "PreToolUse", "tool_name": "Bash"}
            )
            + "\n",
            encoding="utf-8",
        )
        observer = ClaudeHooksObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.RUNNING
        assert events[0].confidence == LifecycleConfidence.HIGH
        assert events[0].detail == "using Bash"

    def test_post_tool_use_sets_running(self, tmp_path):
        log_path = tmp_path / "hooks.jsonl"
        log_path.write_text(
            json.dumps(
                {"hook_event_name": "PostToolUse", "tool_name": "Read"}
            )
            + "\n",
            encoding="utf-8",
        )
        observer = ClaudeHooksObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.RUNNING
        assert events[0].confidence == LifecycleConfidence.HIGH
        assert events[0].detail == "Read done"

    def test_poll_is_incremental(self, tmp_path):
        log_path = tmp_path / "hooks.jsonl"
        log_path.write_text("", encoding="utf-8")
        observer = ClaudeHooksObserver(str(log_path))
        assert observer.poll() == []
        with log_path.open("a", encoding="utf-8") as handle:
            handle.write(json.dumps({"hook_event_name": "TaskCompleted"}) + "\n")
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.COMPLETED
        assert observer.poll() == []


class TestCodexAppServerObserver:
    def test_turn_started_sets_running(self, tmp_path):
        log_path = tmp_path / "events.jsonl"
        log_path.write_text(
            json.dumps({"method": "turn/started"}) + "\n",
            encoding="utf-8",
        )
        observer = CodexAppServerObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.RUNNING
        assert events[0].confidence == LifecycleConfidence.HIGH

    def test_waiting_on_approval_sets_approving(self, tmp_path):
        log_path = tmp_path / "events.jsonl"
        log_path.write_text(
            json.dumps(
                {
                    "method": "thread/status/changed",
                    "params": {"status": "waitingOnApproval"},
                }
            )
            + "\n",
            encoding="utf-8",
        )
        observer = CodexAppServerObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.WAITING_FOR_APPROVAL

    def test_turn_completed_sets_completed(self, tmp_path):
        log_path = tmp_path / "events.jsonl"
        log_path.write_text(
            json.dumps({"method": "turn/completed"}) + "\n",
            encoding="utf-8",
        )
        observer = CodexAppServerObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.COMPLETED

    def test_turn_failed_sets_failed(self, tmp_path):
        log_path = tmp_path / "events.jsonl"
        log_path.write_text(
            json.dumps(
                {
                    "method": "turn/failed",
                    "params": {"error": "tool timeout"},
                }
            )
            + "\n",
            encoding="utf-8",
        )
        observer = CodexAppServerObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.FAILED
        assert "tool timeout" in events[0].detail


class TestGeminiTelemetryObserver:
    def test_activity_event_sets_running(self, tmp_path):
        log_path = tmp_path / "telemetry.jsonl"
        log_path.write_text(
            json.dumps({"event_type": "tool_call"}) + "\n",
            encoding="utf-8",
        )
        observer = GeminiTelemetryObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.RUNNING
        assert events[0].confidence == LifecycleConfidence.MEDIUM

    def test_error_event_sets_failed(self, tmp_path):
        log_path = tmp_path / "telemetry.jsonl"
        log_path.write_text(
            json.dumps({"event_type": "error", "message": "auth failed"}) + "\n",
            encoding="utf-8",
        )
        observer = GeminiTelemetryObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.FAILED
        assert "auth failed" in events[0].detail

    def test_session_end_sets_completed(self, tmp_path):
        log_path = tmp_path / "telemetry.jsonl"
        log_path.write_text(
            json.dumps({"event_type": "session_end"}) + "\n",
            encoding="utf-8",
        )
        observer = GeminiTelemetryObserver(str(log_path))
        events = observer.poll()
        assert len(events) == 1
        assert events[0].status == AgentStatus.COMPLETED



from __future__ import annotations

from dataclasses import dataclass

from textual.app import ComposeResult
from textual.binding import Binding
from textual.widgets import Static

from lazyagent.widgets.confirm_modal import ConfirmModal


@dataclass
class RemoveWorktreeResult:
    """Result returned when the remove worktree modal is confirmed."""

    force: bool


class RemoveWorktreeModal(ConfirmModal):
    """Modal for confirming worktree removal with optional force flag."""

    BINDINGS = [
        Binding("f", "toggle_force", "Toggle force", show=False),
    ]

    def __init__(self, title: str, body: str, **kwargs) -> None:
        super().__init__(title=title, body=body, **kwargs)
        self._force = False

    def _extra_compose(self) -> ComposeResult:
        yield Static(self._force_label(), id="force-toggle")

    def _hint_text(self) -> str:
        return (
            "[bold cyan]y[/bold cyan] yes  "
            "[bold cyan]n[/bold cyan] no  "
            "[bold cyan]f[/bold cyan] toggle force"
        )

    def _force_label(self) -> str:
        marker = "x" if self._force else " "
        return f"\\[{marker}] Force removal"

    def action_confirm(self) -> None:
        self.dismiss(RemoveWorktreeResult(force=self._force))

    def action_deny(self) -> None:
        self.dismiss(None)

    def action_toggle_force(self) -> None:
        self._force = not self._force
        self.query_one("#force-toggle", Static).update(self._force_label())



from __future__ import annotations

import json
import shlex
from lazyagent.agent_observers import (
    ClaudeHooksObserver,
    CodexAppServerObserver,
)
from lazyagent.agent_providers import (
    DEFAULT_AGENT_PROVIDER,
    ObservationMode,
    get_agent_provider,
    normalize_provider_name,
)


class TestNormalizeProviderName:
    def test_defaults_to_claude_for_none(self):
        assert normalize_provider_name(None) == DEFAULT_AGENT_PROVIDER

    def test_normalizes_case_and_whitespace(self):
        assert normalize_provider_name("  CoDeX  ") == "codex"

    def test_normalizes_gemini(self):
        assert normalize_provider_name(" Gemini ") == "gemini"

    def test_invalid_provider_falls_back_to_default(self):
        assert normalize_provider_name("geminix") == DEFAULT_AGENT_PROVIDER


class TestGetAgentProvider:
    def test_claude_uses_hooks_observation(self):
        provider = get_agent_provider("claude")
        assert provider.executable == "claude"
        assert provider.observation_mode == ObservationMode.HOOKS

    def test_codex_uses_own_dangerous_flag(self):
        provider = get_agent_provider("codex")
        assert provider.executable == "codex"
        assert (
            provider.dangerous_flag
            == "--dangerously-bypass-approvals-and-sandbox"
        )
        assert provider.observation_mode == ObservationMode.APP_SERVER
        assert provider.supports_structured_turn_events is True

    def test_gemini_uses_approval_mode_yolo(self):
        provider = get_agent_provider("gemini")
        assert provider.executable == "gemini"
        assert provider.dangerous_flag == "--approval-mode=yolo"
        assert provider.observation_mode == ObservationMode.TELEMETRY

    def test_invalid_provider_returns_default_provider(self):
        provider = get_agent_provider("other")
        assert provider.name == DEFAULT_AGENT_PROVIDER


class TestBuildCommand:
    def test_claude_command_does_not_append_system_prompt(self):
        command = get_agent_provider("claude").build_command("/tmp/wt")
        script = shlex.split(command)[2]
        assert "--append-system-prompt" not in script

    def test_gemini_command_uses_yolo_flag_when_requested(self):
        command = get_agent_provider("gemini").build_command(
            "/tmp/wt",
            skip_permissions=True,
        )
        script = shlex.split(command)[2]
        assert "exec gemini" in script
        assert "--approval-mode=yolo" in script


class TestRuntimeContext:
    def test_build_runtime_context_uses_provider_metadata(self):
        provider = get_agent_provider("claude")
        context = provider.build_runtime_context("/tmp/wt")
        assert context.provider_name == "claude"
        assert context.worktree_path == "/tmp/wt"
        assert context.observation_mode == ObservationMode.HOOKS
        assert "LAZYAGENT_CLAUDE_HOOK_LOG" in context.env_overrides
        assert "CLAUDE_CONFIG_DIR" not in context.env_overrides
        assert "hook_log_path" in context.metadata
        assert "settings_path" in context.metadata

    def test_codex_returns_app_server_observer(self):
        observer = get_agent_provider("codex").create_observer("/tmp/wt")
        assert isinstance(observer, CodexAppServerObserver)

    def test_claude_returns_hooks_observer(self):
        observer = get_agent_provider("claude").create_observer("/tmp/wt")
        assert isinstance(observer, ClaudeHooksObserver)


class TestClaudeHooksSettings:
    def test_settings_file_contains_only_hooks(self):
        provider = get_agent_provider("claude")
        context = provider.build_runtime_context("/tmp/wt")

        settings_path = context.metadata["settings_path"]
        settings = json.loads(open(settings_path, encoding="utf-8").read())

        # Only hooks, no user config leaking in
        assert list(settings.keys()) == ["hooks"]

        # All expected hook events are present
        assert "Notification" in settings["hooks"]
        assert "Stop" in settings["hooks"]
        assert "PreToolUse" in settings["hooks"]
        assert "PostToolUse" in settings["hooks"]
        assert "TaskCompleted" in settings["hooks"]
        assert "SessionEnd" in settings["hooks"]

    def test_build_command_includes_settings_flag(self):
        provider = get_agent_provider("claude")
        context = provider.build_runtime_context("/tmp/wt")
        command = provider.build_command("/tmp/wt", runtime_context=context)
        script = shlex.split(command)[2]
        assert "--settings" in script
        assert context.metadata["settings_path"] in script



from __future__ import annotations

from pathlib import Path
from unittest.mock import MagicMock

import pytest
from textual.widgets import TabbedContent

from lazyagent.app import LazyAgent
from lazyagent.config import Config, WorktreeConfig
from lazyagent.models import AgentState, AgentStatus, GitStatus, WorktreeInfo
from lazyagent.widgets.center_panel import CenterPanel, WorktreePanel
from lazyagent.widgets.create_worktree_modal import CreateWorktreeResult
from lazyagent.widgets.remove_worktree_modal import RemoveWorktreeResult
from lazyagent.widgets.worktree_list import WorktreeList, WorktreeListItem


MAIN_WORKTREE = WorktreeInfo(
    path="/repo",
    head="a" * 40,
    branch="main",
    is_main=True,
    is_bare=False,
)

FEATURE_WORKTREE = WorktreeInfo(
    path="/repo-feature",
    head="b" * 40,
    branch="feature/demo",
    is_main=False,
    is_bare=False,
)

WORKTREES = [MAIN_WORKTREE, FEATURE_WORKTREE]


def _make_app_with_config(
    create: str | None = None,
    remove: str | None = None,
) -> LazyAgent:
    """Create a LazyAgent with custom worktree config and mocked I/O."""
    app = LazyAgent(repo_path="/repo")
    app._repo_root = "/repo"
    app._config = Config(worktree=WorktreeConfig(create=create, remove=remove))
    app._send_to_terminal = MagicMock()
    app.action_focus_terminal = MagicMock()
    app.notify = MagicMock()
    return app


class DummyWorktreeManager:
    def __init__(self, repo_path: str | Path) -> None:
        self.repo_path = Path(repo_path)

    def list(self) -> list[WorktreeInfo]:
        return WORKTREES

    def get_all_git_statuses(
        self, worktrees: list[WorktreeInfo]
    ) -> dict[str, GitStatus]:
        return {
            wt.path: GitStatus(last_commit_subject=f"commit for {wt.name}")
            for wt in worktrees
        }

    @staticmethod
    def get_diff(worktree_path: str) -> str:
        return f"diff for {worktree_path}"

    @staticmethod
    def is_gh_available() -> bool:
        return False


def _patch_app_dependencies(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("lazyagent.app.WorktreeManager", DummyWorktreeManager)
    monkeypatch.setattr("lazyagent.app.load_config", lambda repo_root: Config())
    monkeypatch.setattr(
        WorktreePanel,
        "_try_start_terminal",
        lambda self: None,
    )

    def no_pr_refresh(self) -> None:
        return None

    monkeypatch.setattr(LazyAgent, "_refresh_pr_status", no_pr_refresh)


def _select_worktree(app: LazyAgent, index: int) -> WorktreeInfo:
    worktree_list = app.query_one(WorktreeList)
    item = [
        child
        for child in worktree_list.children
        if isinstance(child, WorktreeListItem)
    ][index]
    worktree_list.index = index
    app.on_list_view_highlighted(WorktreeList.Highlighted(worktree_list, item))
    return item.worktree


@pytest.mark.asyncio
async def test_switching_worktrees_restores_existing_panel_state(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _patch_app_dependencies(monkeypatch)

    app = LazyAgent(repo_path="/repo")

    async with app.run_test() as pilot:
        first = _select_worktree(app, 0)
        await pilot.pause()

        center = app.query_one(CenterPanel)
        first_panel = center.get_panel(first.path)
        assert first_panel is not None

        first_panel.switch_to_tab("diff-tab")
        first_tabs = first_panel.query_one("#agent-tabs", TabbedContent)
        assert first_tabs.active == "diff-tab"

        second = _select_worktree(app, 1)
        await pilot.pause()

        second_panel = center.get_panel(second.path)
        assert second_panel is not None
        assert second_panel is not first_panel

        _select_worktree(app, 0)
        await pilot.pause()

        assert center.get_panel(first.path) is first_panel
        assert first_panel.query_one("#agent-tabs", TabbedContent).active == "diff-tab"


@pytest.mark.asyncio
async def test_ctrl_j_uses_spawn_flow_when_selected_worktree_has_no_agent(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _patch_app_dependencies(monkeypatch)

    app = LazyAgent(repo_path="/repo")

    async with app.run_test() as pilot:
        _select_worktree(app, 1)
        await pilot.pause()

        app.action_spawn_agent = MagicMock()

        await pilot.press("ctrl+j")
        await pilot.pause()

        app.action_spawn_agent.assert_called_once_with()


def test_do_create_worktree_injects_custom_command_into_terminal() -> None:
    app = _make_app_with_config(
        create="create-worktree {branch} {name} {base} {path} {repo}"
    )

    app._do_create_worktree(
        CreateWorktreeResult(branch="feature/demo", base_branch="main")
    )

    app._send_to_terminal.assert_called_once_with(
        "create-worktree feature/demo repo-feature/demo main /repo-feature/demo /repo"
    )
    app.notify.assert_called_once()
    assert "press r to refresh when done" in app.notify.call_args.args[0]


def test_do_remove_worktree_injects_custom_command_with_repo_cd_prefix() -> None:
    app = _make_app_with_config(remove="remove-worktree {name} {path}")

    app._do_remove_worktree(FEATURE_WORKTREE)

    app._send_to_terminal.assert_called_once_with(
        "cd /repo && remove-worktree repo-feature /repo-feature"
    )
    app.action_focus_terminal.assert_called_once_with()


@pytest.mark.parametrize("status", [AgentStatus.RUNNING, AgentStatus.WAITING])
def test_remove_worktree_is_blocked_for_active_agents(status: AgentStatus) -> None:
    app = LazyAgent(repo_path="/repo")
    app._get_selected_worktree = lambda: FEATURE_WORKTREE
    app._get_agent_state = lambda path: AgentState(status=status)
    app.notify = MagicMock()
    app.push_screen = MagicMock()

    app.action_remove_worktree()

    app.push_screen.assert_not_called()
    app.notify.assert_called_once()
    assert "stop it first" in app.notify.call_args.args[0]


def test_remove_worktree_is_blocked_for_main_worktree() -> None:
    app = LazyAgent(repo_path="/repo")
    app._get_selected_worktree = lambda: MAIN_WORKTREE
    app.notify = MagicMock()
    app.push_screen = MagicMock()

    app.action_remove_worktree()

    app.push_screen.assert_not_called()
    app.notify.assert_called_once()
    assert "Cannot remove the main worktree" in app.notify.call_args.args[0]


# --- RemoveWorktreeResult tests ---


def test_remove_worktree_result_defaults() -> None:
    result = RemoveWorktreeResult(force=False)
    assert result.force is False


def test_remove_worktree_result_force() -> None:
    result = RemoveWorktreeResult(force=True)
    assert result.force is True


def test_do_remove_worktree_passes_force_to_builtin() -> None:
    app = _make_app_with_config()
    mock_manager = MagicMock()
    monkeypatch_manager = MagicMock()
    mock_manager.return_value = monkeypatch_manager
    app._load_worktrees = MagicMock()

    import lazyagent.app as app_module

    original = app_module.WorktreeManager
    app_module.WorktreeManager = mock_manager
    try:
        app._do_remove_worktree(FEATURE_WORKTREE, force=True)
        monkeypatch_manager.remove.assert_called_once_with("/repo-feature", force=True)
    finally:
        app_module.WorktreeManager = original


def test_do_remove_worktree_expands_force_placeholder_in_custom_command() -> None:
    app = _make_app_with_config(remove="remove-worktree {name} {path} {force}")

    app._do_remove_worktree(FEATURE_WORKTREE, force=True)

    app._send_to_terminal.assert_called_once_with(
        "cd /repo && remove-worktree repo-feature /repo-feature --force"
    )


def test_do_remove_worktree_custom_without_force() -> None:
    app = _make_app_with_config(remove="remove-worktree {name} {path} {force}")

    app._do_remove_worktree(FEATURE_WORKTREE, force=False)

    app._send_to_terminal.assert_called_once_with(
        "cd /repo && remove-worktree repo-feature /repo-feature"
    )


# --- CreateWorktreeResult extra_options tests ---


def test_create_worktree_result_extra_options_default() -> None:
    result = CreateWorktreeResult(branch="feat", base_branch="main")
    assert result.extra_options == ""


def test_create_worktree_result_extra_options() -> None:
    result = CreateWorktreeResult(
        branch="feat", base_branch="main", extra_options="--no-build"
    )
    assert result.extra_options == "--no-build"


def test_do_create_worktree_expands_extra_placeholder() -> None:
    app = _make_app_with_config(
        create="create-worktree {branch} {name} {base} {path} {repo} {extra}"
    )

    app._do_create_worktree(
        CreateWorktreeResult(
            branch="feature/demo",
            base_branch="main",
            extra_options="--no-build --skip-hooks",
        )
    )

    app._send_to_terminal.assert_called_once_with(
        "create-worktree feature/demo repo-feature/demo main /repo-feature/demo /repo --no-build --skip-hooks"
    )


def test_do_create_worktree_no_extra_options() -> None:
    app = _make_app_with_config(
        create="create-worktree {branch} {name} {base} {path} {repo} {extra}"
    )

    app._do_create_worktree(
        CreateWorktreeResult(branch="feature/demo", base_branch="main")
    )

    app._send_to_terminal.assert_called_once_with(
        "create-worktree feature/demo repo-feature/demo main /repo-feature/demo /repo"
    )



from __future__ import annotations

from textual.message import Message

from lazyagent.models import AgentStatus, LifecycleConfidence


class AgentStatusChanged(Message):
    """Agent status transition (RUNNING, WAITING, POSSIBLY_HANGED)."""

    def __init__(
        self,
        worktree_path: str,
        status: AgentStatus,
        confidence: LifecycleConfidence = LifecycleConfidence.LOW,
        detail: str = "",
    ) -> None:
        super().__init__()
        self.worktree_path = worktree_path
        self.status = status
        self.confidence = confidence
        self.detail = detail


class AgentExited(Message):
    """Claude process disconnected from the pty."""

    def __init__(self, worktree_path: str) -> None:
        super().__init__()
        self.worktree_path = worktree_path



from __future__ import annotations

import json
import os
import shlex
import stat
import tempfile
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path


class ObservationMode(Enum):
    TERMINAL = "terminal"
    HOOKS = "hooks"
    APP_SERVER = "app_server"
    TELEMETRY = "telemetry"

DEFAULT_AGENT_PROVIDER = "claude"

# Vars that the PTY emulator already sets or that may cause issues if overridden.
ENV_SKIP = frozenset({"TERM", "LC_ALL", "HOME", "_"})


@dataclass(frozen=True)
class ProviderRuntimeContext:
    """Spawn-time runtime context for provider command + observability."""

    provider_name: str
    worktree_path: str
    observation_mode: ObservationMode
    env_overrides: dict[str, str] = field(default_factory=dict)
    metadata: dict[str, str] = field(default_factory=dict)


@dataclass(frozen=True)
class AgentProvider:
    """Description of a supported agent CLI."""

    name: str
    executable: str
    dangerous_flag: str
    observation_mode: ObservationMode = ObservationMode.TERMINAL
    supports_structured_turn_events: bool = False
    supports_approval_events: bool = False
    supports_completion_events: bool = False

    def build_command(
        self,
        worktree_path: str,
        skip_permissions: bool = False,
        runtime_context: ProviderRuntimeContext | None = None,
    ) -> str:
        """Build the full shell command used to launch this provider."""
        context = runtime_context or self.build_runtime_context(worktree_path)
        parts = [self.executable]
        if skip_permissions:
            parts.append(self.dangerous_flag)
        if "settings_path" in context.metadata:
            parts.extend(["--settings", context.metadata["settings_path"]])
        inner_cmd = " ".join(shlex.quote(p) for p in parts)
        script = (
            f"{env_exports(context.env_overrides)}"
            f" && cd {shlex.quote(worktree_path)}"
            f" && exec {inner_cmd}"
        )
        return f"bash -c {shlex.quote(script)}"

    def build_runtime_context(self, worktree_path: str) -> ProviderRuntimeContext:
        """Build provider-specific runtime metadata for the spawned session."""
        if self.name == "claude":
            return _build_claude_runtime_context(self, worktree_path)
        if self.name == "codex":
            return _build_codex_runtime_context(self, worktree_path)
        if self.name == "gemini":
            return _build_gemini_runtime_context(self, worktree_path)
        return ProviderRuntimeContext(
            provider_name=self.name,
            worktree_path=worktree_path,
            observation_mode=self.observation_mode,
        )

    def create_observer(self, worktree_path: str):
        """Create the lifecycle observer for this provider."""
        context = self.build_runtime_context(worktree_path)
        return self.create_observer_from_context(context)

    def create_observer_from_context(self, context: ProviderRuntimeContext):
        """Create the lifecycle observer for the prepared runtime context."""
        from lazyagent.agent_observers import (
            ClaudeHooksObserver,
            CodexAppServerObserver,
            CompositeObserver,
            GeminiTelemetryObserver,
            GeminiPromptObserver,
        )

        observers = []
        if self.name == "claude":
            observers.append(
                ClaudeHooksObserver(
                    context.metadata["hook_log_path"],
                    temp_dir=context.metadata["temp_dir"],
                )
            )
        elif self.name == "codex":
            observers.append(
                CodexAppServerObserver(
                    context.metadata["app_server_log_path"],
                    temp_dir=context.metadata["temp_dir"],
                )
            )
        elif self.name == "gemini":
            observers.append(
                GeminiTelemetryObserver(
                    context.metadata["telemetry_log_path"],
                    temp_dir=context.metadata["temp_dir"],
                )
            )
            observers.append(GeminiPromptObserver())

        if len(observers) == 1:
            return observers[0]
        return CompositeObserver(observers)


PROVIDERS = {
    "claude": AgentProvider(
        name="claude",
        executable="claude",
        dangerous_flag="--dangerously-skip-permissions",
        observation_mode=ObservationMode.HOOKS,
        supports_approval_events=True,
        supports_completion_events=True,
    ),
    "codex": AgentProvider(
        name="codex",
        executable="codex",
        dangerous_flag="--dangerously-bypass-approvals-and-sandbox",
        observation_mode=ObservationMode.APP_SERVER,
        supports_structured_turn_events=True,
        supports_approval_events=True,
        supports_completion_events=True,
    ),
    "gemini": AgentProvider(
        name="gemini",
        executable="gemini",
        dangerous_flag="--approval-mode=yolo",
        observation_mode=ObservationMode.TELEMETRY,
        supports_completion_events=True,
    ),
}


def env_exports(extra_env: dict[str, str] | None = None) -> str:
    """Build a shell snippet that restores the parent process environment."""
    merged = {
        key: val
        for key, val in os.environ.items()
        if key not in ENV_SKIP
    }
    if extra_env:
        merged.update(extra_env)

    parts = []
    for key, val in merged.items():
        parts.append(f"{key}={shlex.quote(val)}")
    return "export " + " ".join(parts) if parts else "true"


def normalize_provider_name(provider: str | None) -> str:
    """Normalize a configured provider name to a supported identifier."""
    candidate = (provider or DEFAULT_AGENT_PROVIDER).strip().lower()
    if candidate not in PROVIDERS:
        return DEFAULT_AGENT_PROVIDER
    return candidate


def get_agent_provider(provider: str | None) -> AgentProvider:
    """Return the supported provider for the given config value."""
    return PROVIDERS[normalize_provider_name(provider)]


def _build_gemini_runtime_context(
    provider: AgentProvider,
    worktree_path: str,
) -> ProviderRuntimeContext:
    temp_dir = Path(tempfile.mkdtemp(prefix="lazyagent-gemini-telemetry-"))
    telemetry_log_path = temp_dir / "telemetry.jsonl"

    return ProviderRuntimeContext(
        provider_name=provider.name,
        worktree_path=worktree_path,
        observation_mode=provider.observation_mode,

        env_overrides={
            "GEMINI_TELEMETRY_LOG": str(telemetry_log_path),
        },
        metadata={
            "temp_dir": str(temp_dir),
            "telemetry_log_path": str(telemetry_log_path),
        },
    )


def _build_codex_runtime_context(
    provider: AgentProvider,
    worktree_path: str,
) -> ProviderRuntimeContext:
    temp_dir = Path(tempfile.mkdtemp(prefix="lazyagent-codex-events-"))
    app_server_log_path = temp_dir / "app-server-events.jsonl"

    # NOTE: In a real implementation we might start a sidecar process here
    # or use a Codex-specific flag to pipe events to this file.
    # For now, we prepare the environment so Codex knows where to log.

    return ProviderRuntimeContext(
        provider_name=provider.name,
        worktree_path=worktree_path,
        observation_mode=provider.observation_mode,

        env_overrides={
            "CODEX_APP_SERVER_LOG": str(app_server_log_path),
        },
        metadata={
            "temp_dir": str(temp_dir),
            "app_server_log_path": str(app_server_log_path),
        },
    )


def _build_claude_runtime_context(
    provider: AgentProvider,
    worktree_path: str,
) -> ProviderRuntimeContext:
    temp_dir = Path(tempfile.mkdtemp(prefix="lazyagent-claude-hooks-"))
    hook_log_path = temp_dir / "hook-events.jsonl"
    hook_script_path = temp_dir / "log_hook.py"
    settings_path = temp_dir / "hooks-settings.json"

    hook_script_path.write_text(_claude_hook_script(), encoding="utf-8")
    hook_script_path.chmod(hook_script_path.stat().st_mode | stat.S_IXUSR)

    hook_command = {"type": "command", "command": str(hook_script_path)}

    hooks_settings = {
        "hooks": {
            "Notification": [
                {
                    "matcher": "permission_prompt|idle_prompt|elicitation_dialog",
                    "hooks": [hook_command],
                }
            ],
            "Stop": [{"hooks": [hook_command]}],
            "TaskCompleted": [{"hooks": [hook_command]}],
            "SessionEnd": [{"hooks": [hook_command]}],
            "PreToolUse": [{"hooks": [hook_command]}],
            "PostToolUse": [{"hooks": [hook_command]}],
        }
    }
    settings_path.write_text(json.dumps(hooks_settings), encoding="utf-8")

    return ProviderRuntimeContext(
        provider_name=provider.name,
        worktree_path=worktree_path,
        observation_mode=provider.observation_mode,
        env_overrides={
            "LAZYAGENT_CLAUDE_HOOK_LOG": str(hook_log_path),
        },
        metadata={
            "temp_dir": str(temp_dir),
            "hook_log_path": str(hook_log_path),
            "hook_script_path": str(hook_script_path),
            "settings_path": str(settings_path),
        },
    )


def _claude_hook_script() -> str:
    return """#!/usr/bin/env python3
import json
import os
import sys
from pathlib import Path

payload = sys.stdin.read()
if not payload.strip():
    raise SystemExit(0)

try:
    data = json.loads(payload)
except json.JSONDecodeError:
    raise SystemExit(0)

log_path = os.environ.get("LAZYAGENT_CLAUDE_HOOK_LOG")
if not log_path:
    raise SystemExit(0)

path = Path(log_path)
path.parent.mkdir(parents=True, exist_ok=True)
with path.open("a", encoding="utf-8") as handle:
    handle.write(json.dumps(data) + "\\n")
"""



from __future__ import annotations

import json
import shutil
from dataclasses import dataclass
from pathlib import Path

from lazyagent.models import AgentStatus, LifecycleConfidence


@dataclass(frozen=True)
class AgentLifecycleEvent:
    """Normalized lifecycle update produced by a provider observer."""

    status: AgentStatus
    confidence: LifecycleConfidence
    detail: str = ""


class AgentObserver:
    """Base interface for provider-specific lifecycle observers."""

    def on_process_started(self) -> list[AgentLifecycleEvent]:
        return []

    def on_terminal_output(self, chars: str) -> list[AgentLifecycleEvent]:
        return []

    def on_screen_update(
        self,
        screen_text: str,
        *,
        current_status: AgentStatus,
        current_detail: str = "",
    ) -> list[AgentLifecycleEvent]:
        return []

    def on_disconnect(self) -> list[AgentLifecycleEvent]:
        return []

    def poll(self) -> list[AgentLifecycleEvent]:
        return []

    def cleanup(self) -> None:
        return None


class CompositeObserver(AgentObserver):
    """Observer that combines multiple observation strategies."""

    def __init__(self, observers: list[AgentObserver]) -> None:
        self._observers = observers

    def on_process_started(self) -> list[AgentLifecycleEvent]:
        return _flatten(observer.on_process_started() for observer in self._observers)

    def on_terminal_output(self, chars: str) -> list[AgentLifecycleEvent]:
        return _flatten(
            observer.on_terminal_output(chars) for observer in self._observers
        )

    def on_screen_update(
        self,
        screen_text: str,
        *,
        current_status: AgentStatus,
        current_detail: str = "",
    ) -> list[AgentLifecycleEvent]:
        return _flatten(
            observer.on_screen_update(
                screen_text,
                current_status=current_status,
                current_detail=current_detail,
            )
            for observer in self._observers
        )

    def on_disconnect(self) -> list[AgentLifecycleEvent]:
        return _flatten(observer.on_disconnect() for observer in self._observers)

    def poll(self) -> list[AgentLifecycleEvent]:
        return _flatten(observer.poll() for observer in self._observers)

    def cleanup(self) -> None:
        for observer in self._observers:
            observer.cleanup()


class JsonlTailObserver(AgentObserver):
    """Base observer that incrementally tails a JSONL log file.

    Subclasses implement ``_map_event`` to convert each parsed JSON object
    into an optional ``AgentLifecycleEvent``.
    """

    def __init__(self, log_path: str, temp_dir: str | None = None) -> None:
        self._log_path = Path(log_path)
        self._temp_dir = Path(temp_dir) if temp_dir else None
        self._offset = 0

    def poll(self) -> list[AgentLifecycleEvent]:
        if not self._log_path.exists():
            return []

        events: list[AgentLifecycleEvent] = []
        with self._log_path.open("r", encoding="utf-8") as handle:
            handle.seek(self._offset)
            for line in handle:
                payload = line.strip()
                if not payload:
                    continue
                try:
                    data = json.loads(payload)
                except json.JSONDecodeError:
                    continue
                mapped = self._map_event(data)
                if mapped is not None:
                    events.append(mapped)
            self._offset = handle.tell()
        return events

    def cleanup(self) -> None:
        if self._temp_dir is not None:
            shutil.rmtree(self._temp_dir, ignore_errors=True)

    def _map_event(self, data: dict) -> AgentLifecycleEvent | None:
        raise NotImplementedError


class ClaudeHooksObserver(JsonlTailObserver):
    """Observer that tails Claude hook events written to a JSONL file."""

    def _map_event(self, data: dict) -> AgentLifecycleEvent | None:
        event_name = data.get("hook_event_name")
        if event_name == "Notification":
            notification_type = str(data.get("notification_type", "")).strip().lower()
            if notification_type == "permission_prompt":
                return AgentLifecycleEvent(
                    status=AgentStatus.WAITING_FOR_APPROVAL,
                    confidence=LifecycleConfidence.HIGH,
                    detail="needs approval",
                )
            if notification_type == "idle_prompt":
                return AgentLifecycleEvent(
                    status=AgentStatus.WAITING_FOR_USER,
                    confidence=LifecycleConfidence.HIGH,
                    detail="idle",
                )
            if notification_type == "elicitation_dialog":
                return AgentLifecycleEvent(
                    status=AgentStatus.WAITING_FOR_USER,
                    confidence=LifecycleConfidence.HIGH,
                    detail="asking a question",
                )
            return None
        if event_name == "Stop":
            # In interactive PTY sessions, Stop means Claude finished responding
            # and is waiting for the next user prompt — WAITING is correct here,
            # not COMPLETED (which is reserved for TaskCompleted/SessionEnd).
            return AgentLifecycleEvent(
                status=AgentStatus.WAITING,
                confidence=LifecycleConfidence.HIGH,
                detail="waiting for input",
            )
        if event_name == "TaskCompleted":
            return AgentLifecycleEvent(
                status=AgentStatus.COMPLETED,
                confidence=LifecycleConfidence.HIGH,
                detail="task completed",
            )
        if event_name == "SessionEnd":
            return AgentLifecycleEvent(
                status=AgentStatus.COMPLETED,
                confidence=LifecycleConfidence.HIGH,
                detail="session ended",
            )
        if event_name == "PreToolUse":
            tool_name = data.get("tool_name", "unknown")
            return AgentLifecycleEvent(
                status=AgentStatus.RUNNING,
                confidence=LifecycleConfidence.HIGH,
                detail=f"using {tool_name}",
            )
        if event_name == "PostToolUse":
            tool_name = data.get("tool_name", "unknown")
            return AgentLifecycleEvent(
                status=AgentStatus.RUNNING,
                confidence=LifecycleConfidence.HIGH,
                detail=f"{tool_name} done",
            )
        return None


class CodexAppServerObserver(JsonlTailObserver):
    """Observer that consumes Codex App Server events from a JSONL file."""

    def _map_event(self, data: dict) -> AgentLifecycleEvent | None:
        # Based on documented Codex App Server event shapes
        method = data.get("method")
        params = data.get("params", {})

        if method == "turn/started":
            return AgentLifecycleEvent(
                status=AgentStatus.RUNNING,
                confidence=LifecycleConfidence.HIGH,
                detail="codex turn started",
            )
        if method == "turn/completed":
            return AgentLifecycleEvent(
                status=AgentStatus.COMPLETED,
                confidence=LifecycleConfidence.HIGH,
                detail="codex turn completed",
            )
        if method == "turn/failed":
            return AgentLifecycleEvent(
                status=AgentStatus.FAILED,
                confidence=LifecycleConfidence.HIGH,
                detail=f"codex turn failed: {params.get('error', 'unknown error')}",
            )
        if method == "thread/status/changed":
            status = params.get("status")
            if isinstance(status, dict):
                active_flags = status.get("activeFlags", [])
                if "waitingOnApproval" in active_flags:
                    return AgentLifecycleEvent(
                        status=AgentStatus.WAITING_FOR_APPROVAL,
                        confidence=LifecycleConfidence.HIGH,
                        detail="codex waiting on approval",
                    )
                if "waitingOnUser" in active_flags:
                    return AgentLifecycleEvent(
                        status=AgentStatus.WAITING_FOR_USER,
                        confidence=LifecycleConfidence.HIGH,
                        detail="codex waiting on user",
                    )
            elif status == "waitingOnApproval":
                return AgentLifecycleEvent(
                    status=AgentStatus.WAITING_FOR_APPROVAL,
                    confidence=LifecycleConfidence.HIGH,
                    detail="codex waiting on approval",
                )
            elif status == "waitingOnUser":
                return AgentLifecycleEvent(
                    status=AgentStatus.WAITING_FOR_USER,
                    confidence=LifecycleConfidence.HIGH,
                    detail="codex waiting on user",
                )
        return None


class GeminiTelemetryObserver(JsonlTailObserver):
    """Observer that interprets Gemini telemetry streams from a JSONL file."""

    def _map_event(self, data: dict) -> AgentLifecycleEvent | None:
        event_type = data.get("event_type")

        # Telemetry provides activity signals (with gemini_cli. prefix)
        if event_type in {
            "gemini_cli.api_request",
            "gemini_cli.tool_call",
            "gemini_cli.file.operation",
            "request",  # keep old values as fallback
            "tool_call",
            "file_operation",
        }:
            return AgentLifecycleEvent(
                status=AgentStatus.RUNNING,
                confidence=LifecycleConfidence.MEDIUM,
                detail=f"gemini activity: {event_type}",
            )
        if event_type in {
            "gemini_cli.turn_end",
            "gemini_cli.waiting_for_input",
            "gemini_cli.prompt",
            "gemini_cli.api_response",  # After response, we usually wait
        }:
            return AgentLifecycleEvent(
                status=AgentStatus.WAITING,
                confidence=LifecycleConfidence.HIGH,
                detail=f"gemini {event_type}",
            )
        if event_type in {"gemini_cli.error", "error"}:
            return AgentLifecycleEvent(
                status=AgentStatus.FAILED,
                confidence=LifecycleConfidence.HIGH,
                detail=f"gemini error: {data.get('message', 'unknown')}",
            )
        if event_type in {"gemini_cli.session", "session_end", "gemini_cli.completion"}:
            # session event usually marks lifecycle boundaries
            return AgentLifecycleEvent(
                status=AgentStatus.COMPLETED,
                confidence=LifecycleConfidence.MEDIUM,
                detail="gemini session update",
            )
        return None


class GeminiPromptObserver(AgentObserver):
    """Observer that detects the Gemini CLI prompt (e.g., 'gemini >') on the screen."""

    def on_screen_update(
        self,
        screen_text: str,
        *,
        current_status: AgentStatus,
        current_detail: str = "",
    ) -> list[AgentLifecycleEvent]:
        # Typical gemini prompts end with "> " or "gemini > "
        # We check if any of the last few lines (non-empty) matches this
        lines = [line.strip() for line in screen_text.splitlines() if line.strip()]
        if not lines:
            return []

        last_line = lines[-1]
        # Match "gemini >", "ai >", or just ">" at the end of the line
        if last_line.endswith(">") or " > " in last_line:
            if current_status == AgentStatus.RUNNING:
                return [
                    AgentLifecycleEvent(
                        status=AgentStatus.WAITING,
                        confidence=LifecycleConfidence.MEDIUM,
                        detail="gemini prompt detected on screen",
                    )
                ]
        elif (
            current_status == AgentStatus.WAITING
            and current_detail == "gemini prompt detected on screen"
        ):
            return [
                AgentLifecycleEvent(
                    status=AgentStatus.RUNNING,
                    confidence=LifecycleConfidence.LOW,
                    detail="gemini prompt no longer visible",
                )
            ]
        return []


def _flatten(batches) -> list[AgentLifecycleEvent]:
    events: list[AgentLifecycleEvent] = []
    for batch in batches:
        events.extend(batch)
    return events
